#include "LearnerStage.h"

void LearnerStage(double* popCosts, double* population, double* newPopCosts, double* newPopulation, double* randomDepot, int* pairDepot,
                  double* bounds, double (*objFunc)(double*), int numVars, int popSize, int focusStart, int focusEnd){

    #pragma omp parallel
    {   //Thread-internal variables:
        int partnerFocus = 0;//The individual to interact with (the pairDepot vector should not contain self-interaction)
        int popOffset = 0, bestOffset = 0;//Where to start analyzing the individual "i"
        int partnerOffset = 0, worstOffset = 0;//Where to start analyzing the individual "j"
        int varFocus = 0;//Auxiliary to save the position of the bounds
        double* newIndividual = malloc(sizeof(double)*numVars);//We will assume that there is enough memory by now...
        double newCost = 0.0;//Where to save the value of the individual after the interaction

        #pragma omp for
        for(int i = focusStart; i < focusEnd; i++){
            popOffset = i*numVars;
			partnerFocus = pairDepot[i];//Reading its corresponding partner according to the monolitic generation
            //Deciding the teaching direction:
            partnerOffset = partnerFocus*numVars;
            if(popCosts[i] > popCosts[partnerFocus]){//The individual "partnerFocus" is better
                bestOffset = partnerOffset; worstOffset = popOffset;
            }else{
                bestOffset = popOffset; worstOffset = partnerOffset;
            }
            for(int j = 0; j < numVars; j++){
                varFocus = 2*j;
                newIndividual[j] = population[popOffset + j] + randomDepot[j]*(population[bestOffset + j] - population[worstOffset + j]);
                if(newIndividual[j]<bounds[varFocus]){
                    newIndividual[j] = bounds[varFocus];
                }else if(newIndividual[j]>bounds[varFocus + 1]){
                    newIndividual[j] = bounds[varFocus + 1];
                }
            }

            newCost = objFunc(newIndividual);//Evaluating the new individual
            if(newCost < popCosts[i]){//Selecting the new individual
                CopyIndividual(newIndividual, &newPopulation[popOffset], numVars);
                newPopCosts[i] = newCost;
            }else{//Transferring the same individual
                CopyIndividual(&population[popOffset], &newPopulation[popOffset], numVars);
                newPopCosts[i] = popCosts[i];
            }
        }
        free(newIndividual);
    }//End of the parallel region
}
