#include "MemoryHandler.h"

void InitializeMemory(int numVars, int popSize, double** popCosts, double** population, double** snapPopCosts, double** snapPopulation, double** randomDepot, double** meanDiffs, int** pairDepot){
    //This is an ordered stage to be able to detect the problem and clean-up the previously allocated memory
    *popCosts = malloc(sizeof(double)*popSize);
    if(!(*popCosts)){//Security check
        printf("Error: No memory for the vector of costs");
        exit(EXIT_FAILURE);
    }
    //Initialization of room for the population: a matrix of popSize*numVars of doubles (Inidividuals in rows):
    *population = malloc(sizeof(double)*popSize*numVars);
    if(!(*population)){//Security check
        printf("Error: No memory for the population");
        free(*popCosts);
        exit(EXIT_FAILURE);
    }

    *snapPopCosts = malloc(sizeof(double)*popSize);
    if(!(*snapPopCosts)){//Security check
        printf("Error: No memory for the snapshot vector of costs");
        free(*popCosts);
        free(*population);
        exit(EXIT_FAILURE);
    }

    *snapPopulation = malloc(sizeof(double)*popSize*numVars);
    if(!(*snapPopulation)){//Security check
        printf("Error: No memory for the snapshot population");
        free(*popCosts);
        free(*population);
        free(*snapPopCosts);
        exit(EXIT_FAILURE);
    }

    *randomDepot = malloc(sizeof(double)*2*numVars);//Two times the number of variables because the random values are directly generated for both the teacher and student stages
    if(!(*randomDepot)){//Security check
        printf("Error: No memory for the random depot");
        free(*popCosts);
        free(*population);
        free(*snapPopCosts);
        free(*snapPopulation);
        exit(EXIT_FAILURE);
    }

    *meanDiffs = malloc(sizeof(double)*numVars);
    if(!(*meanDiffs)){
        printf("Error: No memory for the difference means vector");
        free(*popCosts);
        free(*population);
        free(*snapPopCosts);
        free(*snapPopulation);
        free(*randomDepot);
        exit(EXIT_FAILURE);
    }
    
    *pairDepot = malloc(sizeof(int)*popSize);
    if(!(*pairDepot)){
		printf("Error: No memory for the pairing vector");
        free(*popCosts);
        free(*population);
        free(*snapPopCosts);
        free(*snapPopulation);
        free(*randomDepot);
        free(*meanDiffs);
        exit(EXIT_FAILURE);
	}
}

void CopyIndividual(double* individual, double* outSolution, int numVars){
    for(int i=0; i<numVars; i++){
        outSolution[i] = individual[i];
    }
}

void InitializeRandomVector(double* randomDepot, int length, int* pairDepot, int popSize){
    for(int i = 0; i<length; i++){//Subject factors for both teacher and student's stages:
        randomDepot[i] = (double) rand()/RAND_MAX; //NOTE: rand()/RAND_MAX should be a real value between 0 and RAND_MAX... so by dividing and casting we turn it to decimal
    }//---------------------------
    int goodPair = 0, pair =0;
    for(int i = 0; i<popSize; i++){//Pairs for the student's stage
		goodPair = 0;
		while(!goodPair){
			pair = rand()%popSize;
			if(pair!=i){//No self-pairing
				goodPair = 1;
			}
		}
		pairDepot[i] = pair;
	}
}

void ReadMemory(double* vec, int length){
    for(int i = 0; i<length; i++){
        printf("[%d] -> %.16lf\n", i, vec[i]);
    }
    printf("--------_End_Of_The_Vector_--------\n\n");
}

void CleanMemory(double* popCosts, double* population, double* newPopCosts, double* newPopulation, double* randomDepot, double* meanDiffs, int* pairDepot){
    free(popCosts);
    free(population);
    free(newPopCosts);
    free(newPopulation);
    free(randomDepot);
    free(meanDiffs);
    free(pairDepot);
}
