#include "TLBOCore.h"

//This is a sequential version what includes important notes about developing a full hybrid version of the procedure (MPI processes with internal threads)

double TLBOSearch(double* outSolution, int numVars, double (*objFunc)(double*), int popSize, int numCycles, double* bounds, int focusStart, int focusEnd){
    //PRELIMINARY ACTIVITY-> RESERVING ALL THE MEMORY THAT WILL BE REQUIRED ALONG THE PROCEDURE (No matter which process you are)
    double* popCosts = 0;//Where to save the value of every individual (student)
    double* population = 0;//Where to save the population of students in ROW-MAJOR order (every student is a row)
    double* swapPop = 0; double* swapCosts = 0;//Auxiliary containers to swap between the two branches of population
    double* newPopCosts = 0;//AUXILIARY memory to be able to change the population at the student stage without being affected by the changes
    double* newPopulation = 0;//AUXILIARY memory to be able to change the population at the student stage without being affected by the changes
    double* randomDepot = 0;//Where to save the random numbers generated by the master for both the Teacher and Learner stages (as every MPI-process should have its own random seed for removing duplicates)
    double* meanDiffs = 0;//Where to save the difference means for the teacher stage
    int* pairDepot = 0;//Where to save the pairs for student's stage <Monolitic generation with no self-pairing>
    int TF = 0;//The teaching factor for a certain cycle (what should be computed and shared by the master as every MPI-Process would have a different random seed)

    InitializeMemory(numVars, popSize, &popCosts, &population, &newPopCosts, &newPopulation, &randomDepot, &meanDiffs, &pairDepot);
    //At this point the procedure should be able to be started:

    //STEP_2: Initialize the vector of costs and the population. MONOLITIC GENERATION -> ONLY THE MASTER (WHICH MAY USE INTERNAL THREADS THOUGH)
    #ifndef USE_MPI //focusStart SHOULD BE 0 AND focusEnd SHOULD BE popSize
		Initialize_Population(popCosts, population, objFunc, focusStart, focusEnd, numVars, popSize, bounds);//Initializing the part of the population <MONOLITIC>
    #else
		//Define the overall data distribution (Allgather's...):
        int rank, teamSize, duplicatesFound;
        MPI_Comm_size( MPI_COMM_WORLD, &teamSize );//Total number of MPI processes
        MPI_Comm_rank( MPI_COMM_WORLD, &rank );//Who am I?
        int* counts = malloc(sizeof(int)*teamSize); //How many elements per process? <cost-related>
        int* disps = malloc(sizeof(int)*teamSize);//Where to start saving the data of the process "i" in the array/buffer <cost-related>
        int* popCounts = malloc(sizeof(int)*teamSize); //How many elements per process? <population-related>
        int* popDisps = malloc(sizeof(int)*teamSize);//Where to start saving the data of the process "i" in the array/buffer <population-related>
        int chunkSize = popSize/teamSize;
        int chunkOffset = popSize % teamSize;
        counts[0] = chunkSize + (chunkOffset>0);    popCounts[0] = counts[0]*numVars;
        disps[0] = 0;                               popDisps[0] = 0;//The start of both vectors
        for(int i = 1; i< teamSize; i++){//i == the rank of the "i" process
            disps[i] = i*chunkSize + ((i>=chunkOffset)?chunkOffset:i);//Where should the information of the "i" process started to be saved? <Costs vector>
            counts[i] = chunkSize + (i<chunkOffset);//How many elements will be altered by the "i" process?
            popCounts[i] = counts[i]*numVars;//Equivalent to counts but related to the population
            popDisps[i] = disps[i]*numVars;//Equivalent to disps but related to the population
        }
        if(rank==0){
			Initialize_Population(popCosts, population, objFunc, 0, popSize, numVars, popSize, bounds);//Initializing the part of the population <MONOLITIC>
		}
		MPI_Bcast(popCosts, popSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);//THE MASTER WILL SEND THE POPULATION COSTS TO EVERY ONE
		MPI_Bcast(population, numVars*popSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);//THE MASTER WILL SEND THE INDIVIDUALS TO EVERY ONE
    #endif
    
    for(int i = 0; i<numCycles; i++){//MAIN LOOP: STEPS 3 -> 4 -> 5 -> 3 ... <CYCLES SHOULD BE EXECUTED BY ALL PROCESSES>
        #ifndef USE_MPI
            TF = rand()%2 + 1;//A random integer for the teacher stage which can be 1 OR 2 <expected to be done by the master process>
            InitializeRandomVector(randomDepot, 2*numVars, pairDepot, popSize);//The random values for the teacher and learner stages-> a random value per dimension at every case -> 2*numVars values
        #else //MPI-version
			if(rank==0){//MONOLITIC RANDOM-VALUES GENERATION -> IT KEEPS THE PROCEDURE 100% CONSISTENT WITH THE SEQUENTIAL ONE, JUST TRIES TO SHARE OBJECTIVE FUNCTION EVALUATIONS
				TF = rand()%2 + 1;//A random integer for the teacher stage which can be 1 OR 2 <expected to be done by the master process>
				InitializeRandomVector(randomDepot, 2*numVars, pairDepot, popSize);//The random values for the teacher and learner stages-> a random value per dimension at every case -> 2*numVars values
			}
            MPI_Bcast(&TF, 1, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the teaching factor of the cycle
            MPI_Bcast(randomDepot, 2*numVars, MPI_DOUBLE, 0, MPI_COMM_WORLD);//Sharing the random depot vector
            MPI_Bcast(pairDepot, popSize, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the depot of pairs to evaluate at the student's stage
            ///<TF, randomDepot and parDepot should be shared by the master process with the other ones. THIS APPROACH GRANTS FULLY CONSISTENCY WITH THE PURE SEQUENTIAL VERSION>
        #endif // USE_MPI

        //STEP 3: Teacher stage <Internal threads by OpenMP at every MPI-node> // FocusStart and focusEnd determines where should I write as an MPI-process (they should take all the pop. in the sequential ver.)
        TeacherStage(popCosts, population, randomDepot, meanDiffs, bounds, objFunc, TF, numVars, popSize, focusStart, focusEnd);
        #ifdef USE_MPI
            MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, popCosts, counts, disps, MPI_DOUBLE, MPI_COMM_WORLD);//See: https://www.mpi-forum.org/docs/mpi-2.0/mpi-20-html/node145.htm
            MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, population, popCounts, popDisps, MPI_DOUBLE, MPI_COMM_WORLD);//See: http://stackoverflow.com/questions/15951988/understanding-mpi-allgatherv-in-plain-english
            ///<Population and cost vectors must be synchronized here in the parallel MPI-version> (Allgatherv)
        #endif //USE_MPI

        //STEP 4: Learner stage <Internal threads by OpenMP at every MPI-node>
        LearnerStage(popCosts, population, newPopCosts, newPopulation, &(randomDepot[numVars]), pairDepot, bounds, objFunc, numVars, popSize, focusStart, focusEnd);
        #ifdef USE_MPI
            MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, newPopCosts, counts, disps, MPI_DOUBLE, MPI_COMM_WORLD);//See: https://www.mpi-forum.org/docs/mpi-2.0/mpi-20-html/node145.htm
            MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, newPopulation, popCounts, popDisps, MPI_DOUBLE, MPI_COMM_WORLD);//See: http://stackoverflow.com/questions/15951988/understanding-mpi-allgatherv-in-plain-english
            ///<Population and cost vectors (THE NEW* ONES) must be synchronized here in the parallel MPI-version> (Allgatherv)
        #endif //USE_MPI

        //UNOFFICIAL STEP: Removing duplicates individuals <SEQUENTIAL & CENTRALIZED NOW>: Should it be included... identical values are not likely to exist and this stage is no explained in the original description?
        #ifdef USE_MPI
            if(rank==0){//This is the master process
                duplicatesFound = RemoveDuplicates(newPopCosts, newPopulation, bounds, objFunc, numVars, 0, popSize);//After the learning stage, the current population should be saved in "newPopulation"
            }
	    MPI_Bcast(&duplicatesFound, 1, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the necessity of updating the population
	    if(duplicatesFound){//ADVICE FOR IMPROVEMENT: For relatively probable but not generalized collision environments only changed values should be shared... (Coming in future releases)
		//Let the master share the final results of the cycle before swapping
		MPI_Bcast(newPopCosts, popSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);
                MPI_Bcast(newPopulation, popSize*numVars, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	    }
            ///<Population and cost vectors (THE NEW* ONES) must be distributed here in the parallel MPI-version> (BCAST as it is done by the master process)
        #else //Plain no-MPI version
            RemoveDuplicates(newPopCosts, newPopulation, bounds, objFunc, numVars, 0, popSize);//After the learning stage, the current population should be saved in "newPopulation"
        #endif //USE_MPI

        //SWAPPING <The focus of the population sets should be swapped as the final population is in newPopulation after the cycle>
        swapPop = population; population = newPopulation; newPopulation = swapPop;
        swapCosts = popCosts; popCosts = newPopCosts; newPopCosts = swapCosts;
    }
    #ifdef USE_MPI 
		free(counts); free(disps); free(popCounts); free(popDisps); //Freeing-up the memory related to the specific "allgatherv" operations
		double bestValue = -1.0;//Every MPI process will return -1.0  when it is not the master		
		if(rank==0){
			int bestIndex = LookForTheBest(popCosts, popSize);
			CopyIndividual(&population[bestIndex*numVars], outSolution, numVars);//Origin, destination, length
			bestValue = popCosts[bestIndex];
		}
	#else
		int bestIndex = LookForTheBest(popCosts, popSize);
		CopyIndividual(&population[bestIndex*numVars], outSolution, numVars);//Origin, destination, length
		double bestValue = popCosts[bestIndex];
    #endif // USE_MPI
    CleanMemory(popCosts, population, newPopCosts, newPopulation, randomDepot, meanDiffs, pairDepot);//End of the search: Cleaning the reserved memory: <DONE BY EVERY MPI PROCESS>    
    return bestValue;//The value of the best individual that has been found
}