#include "TeacherStage.h"

void TeacherStage(double* popCosts, double* population, double* randomDepot, double* meanDiffs, double* bounds,
                  double (*objFunc)(double*), int TF, int numVars, int popSize, int focusStart, int focusEnd){
    //Reset of values from previous cycles: <Computed by everyone, expected to be better than wait and share...>
    for(int i = 0; i < numVars; i++){
		meanDiffs[i] = 0.0;
    }
    //Computing the mean by columns (average per subject). for(int i = 0; i < popSize; i++){ for(int j = 0; j < numVars; j++){ mediaNormal[j] += (population[i*numVars + j] / popSize); }}
    for(int i = 0; i<popSize; i++){//Fused option: for(int x = 0; x < popSize*numVars; x++){ meanDiffs[ x % numVars ] += (population[x] / popSize); }//Discarded after benchmarking in my MSI
		for(int j = 0; j < numVars; j++){
			meanDiffs[j] += (population[i*numVars + j] / popSize);//Saving the average
        }
    }
    int teacherIndex = LookForTheBest(popCosts, popSize);//Selecting the teacher
    int popOffset = numVars*teacherIndex;//Where to look for the teacher
    //Updating the meanDiffs vector according to the teacher:
    for(int i = 0; i < numVars; i++){
		meanDiffs[i] = randomDepot[i]*(population[popOffset + i] - TF*meanDiffs[i]);
    }
    ///The previous steps may not require deploying threads... UNDER BENCHMARKING...
    ///The previous steps are assumed to be computed by all MPI processes even though it should be the same information just to reduce interactions as it should not be a costful step
    #pragma omp parallel private(popOffset) //Every thread will need a private copy to write its "popOffset"
    {
	    //As we have the information of the "meanDiffs" we can compute the new values an update the population (only if they are better than the previous one)
	    double* newIndividual = malloc(sizeof(double)*numVars);//Where to save the new candidate individuals. This step should be in a parallel region of OpenMP in the parallel version
	    double newCost = 0.0;                                   //We will assume that there is enough memory by now...
	    int varFocus = 0;//All these values are thread-local
		
	    #pragma omp for
	    for(int i = focusStart; i <focusEnd; i++){//As an MPI process, I will work only over my subset of the population
		    popOffset = i*numVars;
			for(int j = 0; j < numVars; j++){
				newIndividual[j] = population[popOffset + j] + meanDiffs[j];
				varFocus = 2*j;//Where to look for the limits at the bound array
				//Let's ensure that the solution is still in the search-space
				if(newIndividual[j] < bounds[varFocus]){//Less than the lower bound <FIX LIMITS PROCEDURE>
					newIndividual[j] = bounds[varFocus];
				}else if(newIndividual[j] > bounds[varFocus+1]){//Greater than the upper bound
					newIndividual[j] = bounds[varFocus+1];
				}
			}
			newCost = objFunc(newIndividual);//Evaluating the new individual after its creation
			if(newCost<popCosts[i]){//The new one is better than the previous student... let's update it
				popCosts[i] = newCost;
				CopyIndividual(newIndividual, &population[popOffset], numVars);
			}//It will be discarded otherwise
		}
		free(newIndividual);//This step should be in a parallel region of OpenMP in the parallel version
	}//End of the parallel region
}